/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1996 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	cfi.c (Common Filter Interface)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	960130	created
	960203	renamed from approxy (application protocol translation library)
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>

extern int simple_relayf();
extern char *RFC822_readHeader();
extern char *findField();
extern char *nextField();
extern char *matchFields();
extern char *getFieldValue();
extern int copyfile1();
extern FILE *TMPFILE();
extern char *malloc();
extern char *strstr();
extern char *strdup();
#define FTOCL_FIELD	"X-Request"

/*
 * ARGUMENTS
 *
 *	in	 input
 *	out	 output
 *	conninfo information about the client/server connection
 *	convspec specification of conversion
 *
 * RETURN value
 *
 *	-1 error
 *	 0 no translation
 *	 1 translated
 */
cfi(isresp,in,out,conninfo,convspec)
	FILE *in,*out;
	char *conninfo,*convspec;
{	char statline[1024];
	char *head;
	FILE *tmp;
	char clproto[64];
	int withhead;
	int body;

	if( getFieldValue(conninfo,"Client-Protocol",clproto) == 0 )
		strcpy(clproto,"http");

sv1log("## CFI/%s ##\n",clproto);

	if( strcaseeq(clproto,"nntp")
	 || strcaseeq(clproto,"smtp")
	 || strcaseeq(clproto,"pop") )
	{
		filter_msgs(isresp,in,out,conninfo,convspec);
		return 0;
	}

	if(!strcaseeq(clproto,"http") ){
		if( non_http_conv(in,out,conninfo,convspec,clproto) ){
			sv1log("## NON-MIME-RELAY\n");
			return 1;
		}else{
			sv1log("## SIMPLE-RELAY\n");
			simple_relayf(in,out);
			return 0;
		}
	}

	/*
	 *	HTTP
	 */
	withhead = 1;
	body = 1;
	if( isresp ){
		/*
		 *	Pass through non-HTTP stream
		 */
		if( fPollIn(in,0) < 0 )
			exit(-1);
		if( fread(statline,1,5,in) <= 0 )
			return -1;

		if( strncmp(statline,"HTTP/",5) != 0 ){
			write(fileno(out),statline,5);
			simple_relayf(in,out);
			return 0;
		}
		/*
		 *	Read status line and response header
		 */
		if( fgets(statline+5,sizeof(statline),in) == NULL )
			return -1;
	}else{
		fgets(statline,sizeof(statline),in);
		if( strncmp(statline,"GET",3) == 0 )
			body = 0;
		if( strstr(statline,"HTTP/") == NULL ){
			withhead = 0;
			body = 0;
		}
	}
	if( withhead )
		head = RFC822_readHeader(in,0);
	else	head = "";

	/*
	 *	pass through if not to be filterd
	 */
	if( http_conv(in,out,conninfo,convspec,statline,head,body) == 0 ){
		fputs(statline,out);
		removeFields(head,FTOCL_FIELD,1);
		fputs(head,out);
		simple_relayf(in,out);
		return 0;
	}
	return 1;
}

char *searchSpec(conninfo,specs,statline,head)
	char *conninfo,*specs,*statline,*head;
{	char ictype[1024],iagent[1024],iserver[1024],iencode[1024];
	char request[2048],imethod[256],iurl[2048],iver[256];
	char from[256];
	char orequest[2048],iourl[2048];
	char rver[256],rstat[256];
	char clhost[256],svhost[256];
	int top;
	char *spec1,*np;

	orequest[0] = iourl[0] = 0;
	getFieldValue(head,"X-Request-Original",orequest);
	sscanf(orequest,"%*s %s",iourl);

	request[0] = 0;
	getFieldValue(head,"X-Request",request);
	sscanf(request,"%s %s %s",imethod,iurl,iver);
	getFieldValue(head,"X-Request-User-Agent",iagent);

	sscanf(statline,"%s %s",rver,rstat);
	getFieldValue(head,"Server",iserver);
	getFieldValue(head,"Content-Type",ictype);
	getFieldValue(head,"Content-Encoding",iencode);
	getFieldValue(conninfo,"Client-Host",clhost);
	svhost[0] = 0;

	getFieldValue(head,"From",from);
	
	spec1 = specs;
	for( spec1 = specs; *spec1; spec1 = np ){
		top = 1;
		for( np = spec1; *np; np++ ){
			if( top ){
				if( *np == '.' && (np[1]=='\r' || np[1] == '\n')
				 || *np == '-' && np[1]=='-' && (np[2]=='\r'||np[2]=='\n')
				){
					if( *np == '-' )
						np += 2;
					else	np += 1;

					while( np[1] == '\r' || np[1] == '\n' )
						np++;
					*np++ = 0;
					break;
				}
			}
			top = (*np == '\n');
		}

		while( isspace(*spec1) )
			spec1++;
		if( *spec1 == 0 )
			continue;

		sv1vlog("////////////////////////////\n%s\n",spec1);

		if( !matchFields(spec1,"User-Agent",iagent) ) continue;
		if( !matchFields(spec1,"Server",iserver) ) continue;
		if( !matchFields(spec1,"Content-Type",ictype) ) continue;
		if( !matchFields(spec1,"Content-Encoding",iencode) ) continue;
		if( !matchFields(spec1,"Req-Method",imethod) ) continue;
		if( !matchFields(spec1,"Req-Url",iurl) ) continue;
		if( !matchFields(spec1,"Req-Version",iver) ) continue;
		if( !matchFields(spec1,"Req-Url-Orig",iourl) ) continue;
		if( !matchFields(spec1,"Res-Version",rver) ) continue;
		if( !matchFields(spec1,"Res-Status",rstat) ) continue;
		if( !matchFields(spec1,"Client-Host",clhost) ) continue;
		if( !matchFields(spec1,"Server-Host",svhost) ) continue;
		if( !matchFields(spec1,"From",from) ) continue;

		sv1vlog("///////// MATCHED //////////\n");
		return spec1;
	}
	return NULL;
}

static char *OPgetFieldValue(spec,prefix,fname,value)
	char *spec,*prefix,*fname,*value;
{	char *vp,field[256];

	sprintf(field,"%s/%s",prefix,fname);
	if( vp = getFieldValue(spec,field,value) )
		return vp;

	sprintf(field,"%s-%s",prefix,fname); /* obsolete ... */
	if( vp = getFieldValue(spec,field,value) )
		return vp;

	return NULL;
}

static http_conv(in,out,conninfo,convspec,statline,head,body)
	FILE *in,*out;
	char *conninfo,*convspec,*statline,*head;
{	char *spec;
	char octype[1024],ochset[256],oetype[256],ostatus[256];
	char oheader[1024],prefix[1024],postfix[1024];
	char filter[1024],cgi[1024];
	FILE *intmp,*tmp;
	char *xhead;
	int bodyoff;
	char *ohead,*nhead;

	spec = searchSpec(conninfo,convspec,statline,head);
	if( spec == NULL )
		return 0;

	/*
	 *	Rewrite Content-Type field
	 */
	if( OPgetFieldValue(spec,"Output","Content-Type",octype) )
		replaceFieldValue(head,"Content-Type",octype);

	if( OPgetFieldValue(spec,"Output","Charset",ochset) )
		replace_charset(head,ochset);

	if( OPgetFieldValue(spec,"Output","Content-Encoding",oetype) )
		replaceFieldValue(head,"Content-Encoding",oetype);

	filterFields(spec,head);

	tmp = TMPFILE("CFI-CONV");

	ohead = nhead = head;
	if( getFieldValue(spec,"Header/Filter",filter) ){
		FILE *htmp;
		int nhsize;

		htmp = TMPFILE("CFI-Header/Filter");
		fputs(head,htmp); fflush(htmp); fseek(htmp,0,0);
		System(filter,htmp,tmp);
		fclose(htmp);
		fflush(tmp); fseek(tmp,0,0);
		nhsize = file_size(fileno(tmp));
		nhead = malloc(nhsize+2);
		fread(nhead,1,nhsize,tmp);
		nhead[nhsize] = 0;
		fseek(tmp,0,0);
		head = nhead;
	}

	intmp = NULL; 
	xhead = NULL;

	if( getFieldValue(spec,"CGI",cgi) ){
		char oreq[4096],req[4096];

		intmp = TMPFILE("CFI-CGI");
		getFieldValue(head,"X-Request-Original",oreq);
		getFieldValue(head,"X-Request",req);

		system_CGI(conninfo,oreq,req,head,cgi,in,intmp);
		fseek(intmp,0,0);
		fgets(statline,sizeof(statline),intmp);
		head = xhead = RFC822_readHeader(intmp,0);
		in = intmp;
	}

	if( OPgetFieldValue(spec,"Output","Status",ostatus) )
		fprintf(tmp,"%s %s\r\n","HTTP/1.0",ostatus);
	else	fputs(statline,tmp);

	if( getFieldValue(spec,"Output-Header",oheader) ){
		fputs(oheader,tmp);
		fputs("\r\n",tmp);
	}

	removeFields(head,FTOCL_FIELD,1);
	fputs(head,tmp);
	if( getFieldValue(spec,"Output-Prefix",prefix) )
		fputs(prefix,tmp);
	fflush(tmp);
	bodyoff = ftell(tmp);

	/*
	 *	Translate
	 */
	if( body ){
		if( getFieldValue(spec,"Filter",filter) )
			System(filter,in,tmp);
		else	copyfile1(in,tmp);
	}
	if( getFieldValue(spec,"Output-Postfix",postfix) )
		fputs(postfix,tmp);

/*
	fflush(tmp);
	fseek(tmp,bodyoff,0);
	if( fgets(statline,sizeof(statline),tmp) != NULL ){
		if( strncmp(statline,"CFI/",4) == 0 ){
			char *head;

			sv1log("CFI-CONTROL: %s\n",statline);
			if( head = RFC822_readHeader(tmp,0) ){
				sv1log("CFI-CONTROL: %s\n",head);
				free(head);
			}
		}
	}
*/

	fseek(tmp,0,0);
	HTTP_putMIMEmsg(NULL,tmp,out);
	fflush(out);
	fclose(tmp);
	if( intmp ){
		fclose(intmp);
		free(xhead);
	}
	return 1;
}

non_http_conv(in,out,conninfo,convspec,clproto)
	FILE *in,*out;
	char *conninfo,*convspec,*clproto;
{	char *spec,filter[1024];

	spec = searchSpec(conninfo,convspec,"",conninfo);
	if( spec == NULL )
		return 0;

	if( getFieldValue(spec,"Filter",filter) )
		System(filter,in,out);
	else	simple_relayf(in,out);
	return 1;
}

System(command,in,out)
	char *command;
	FILE *in,*out;
{
	if( 0 <= builtin_filter(command,in,out) )
		return;

	systemFilter(command,in,out);
}

extern char **environ;
system_CGI(conninfo,oreq,req,head,cgi,in,out)
	char *conninfo,*oreq,*req,*head,*cgi;
	FILE *in,*out;
{	char *ev[128],eb[0x10000];
	FILE *tmp;
	char *xhead,*dp;
	char url[4096],ourl[4096];

	xhead = strdup(head);
	scan_http_request(oreq,NULL,ourl,NULL);
	scan_http_request(req,NULL,url,NULL);
	cgi_makeenv(conninfo,req,xhead,ourl,url,"","",NULL,ev,eb);
	environ = ev;

	tmp = TMPFILE("CFI-system_CGI");
	System(cgi,in,tmp);
	fseek(tmp,0,0);
	if( dp = strstr(head,"\r\n\r\n") )
		dp[2] = 0;
	if( dp = strstr(head,"\n\n") )
		dp[1] = 0;
	cgi_response(NULL,head,tmp,out);
	fclose(tmp);

	free(xhead);
}


extern char *strSeekEOH();
putMESSAGEline(fp,type,comment)
	FILE *fp;
	char *type,*comment;
{
	return fprintf(fp,"--MESSAGE/%s %s\r\n",type,comment);
}
getMESSAGEline(line,type,comment)
	char *line,*type,*comment;
{
	if( strncmp(line,"--MESSAGE/",10) == 0 )
		return sscanf(line,"--MESSAGE/%s %[^\r\n]",type,comment);
	else	return 0;
}

/*    FILTER_MSGS
 *	Filter a series of mesages which start with a line headed by
 *	  "--MESSAGE/line"
 *	  "--MESSAGE/head"
 *	  "--MESSAGE/body"
 *	  "--MESSAGE/mime"
 *	and end with a line
 *	  "." CRLF
 */

static char *NOHEAD = "";

filter_msgs(isresp,in,out,conninfo,convspec)
	FILE *in,*out;
	char *conninfo,*convspec;
{	FILE *tmp;
	char msgstat[1024];
	char statline[1024];
	char type[128],comment[128];
	char *head;
	int body;
	char *eoh,eohb[1024],endline[1024];

	tmp = TMPFILE("CFI-MESSAGES");

sv1log("## CFI/MSGS ##\n");

	for(;;){
		if( fPollIn(in,10) == 0 )
			fflush(out);
		if( fgets(msgstat,sizeof(msgstat),in) == NULL )
			break;

sv1log("CFI/MSGS ---- %s",msgstat);

		/*
		 *	out of scoope of the filter
		 */
		if( getMESSAGEline(msgstat,type,comment) == 0 ){
			fputs(msgstat,out);
			continue;
		}
		/*
		 *	only status line
		 */
		if( strcmp(type,"line") == 0 ){
			fgets(statline,sizeof(statline),in);
			fputs(statline,out);
			continue;
		}

		/*
		 *	status line and [head][body].CRLF
		 */
		fgets(statline,sizeof(statline),in);
		if( strcmp(type,"head") == 0 ){
			head = RFC822_readHeader(in,1);
			body = 0;
		}else
		if( strcmp(type,"mime") == 0 ){
			head = RFC822_readHeader(in,1);
			body = 1;
		}else
		if( strcmp(type,"body") == 0 ){
			head = NOHEAD;
			body = 1;
		}else{
sv1log("CFI/MSG ---- unknown type: %s\n",type);
			exit(1);
		}
		eoh = strSeekEOH(head);
		if( *eoh == '.' ){
			strcpy(eohb,eoh);
			*eoh = 0;
		}else	eohb[0] = 0;

		endline[0] = 0;
		fseek(tmp,0,0);
		if( body ){
			relayRESPBODY(in,tmp,endline,sizeof(endline));
			fflush(tmp);
			ftruncate(fileno(tmp),ftell(tmp));
			fseek(tmp,0,0);
		}else{
			ftruncate(fileno(tmp),0);
		}
		http_conv(tmp,out,conninfo,convspec,statline,head,body);
		fputs(eohb,out);
		fputs(endline,out);

		if( head != NOHEAD )
			free(head);
	}
}
