/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1995 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	reshost.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	950817	created
//////////////////////////////////////////////////////////////////////#*/

#include <stdio.h>
extern char *strdup();
extern char *strpbrk();
extern char *getenv();
extern char *strchr();
extern char *inet_ntoaV4();

#include "dns.h"
int DNS_debug;

#include "vsocket.h"

#if defined(sgi) || defined(__RES) && (19931104 <= __RES)
#define _RSTATE 1
#else
#define _RSTATE 0
#endif

#ifdef RES_STATE
#define RSTATE !_RSTATE
#else
#define RSTATE  _RSTATE
#endif

#if RSTATE
typedef struct __res_state State;
#else
typedef struct state State;
#endif

#if defined(ultrix)
#define NSLIST(sx)	ns_list[sx].addr
#else
#define NSLIST(sx)	nsaddr_list[sx]
#endif

State _res = {0};
int MIN_ABSNDOTS = 1;

#define MAX_SORTLIST	16
#define ADDRLEN	4
typedef struct {
	unsigned char	addr[ADDRLEN];
	unsigned char	mask[ADDRLEN];
} AddrMask;
static AddrMask sort_list[MAX_SORTLIST];
static int Nsortlist;

#ifndef RSLV_CONF
#define RSLV_CONF	"/etc/resolv.conf"
#endif

#ifndef HOSTSFILE
#define HOSTSFILE	"/etc/hosts"
#endif

#ifndef NISMAP_NAME
#define NISMAP_NAME	"hosts.byname"
#endif

#ifndef NISMAP_ADDR
#define NISMAP_ADDR	"hosts.byaddr"
#endif

#ifndef RSLV_ORDER
#define RSLV_ORDER	"FND"
#endif

char *_RSLV_CONF = RSLV_CONF;
char *_HOSTSFILE = HOSTSFILE;
char *_NISMAP_NAME = NISMAP_NAME;
char *_NISMAP_ADDR = NISMAP_ADDR;

#define RESOLVERS_SIZ 512
char resolvers[RESOLVERS_SIZ] = RSLV_ORDER;

char *RES_VERIFY;

RES_ns1(res,ns,domain)
	State *res;
	char *ns,*domain;
{	SIN sin;
	int nsx;
	int nsin;
	struct hostent *hp;
	char *saddr;

	nsx = -1;
	if( strcmp(domain,RES_NSDOM0) == 0 ){
		/* reserver the position */
		if( res->nscount < MAXNS ){
			nsx = res->nscount++;
			bzero(&res->NSLIST(nsx),sizeof(SIN));
		}
	}

	bzero(&sin,sizeof(SIN));
	sin.sin_addr.s_addr = 0;
	if( (nsin = inet_addrV4(ns)) != -1  )
		bcopy(&nsin,&sin.sin_addr,4);
	else
	if( hp = gethostbyname(ns) )
		bcopy(hp->h_addr_list[0],&sin.sin_addr,4);

	if( sin.sin_addr.s_addr == 0 ){
		/* remove the reserved slot */
		if( 0 <= nsx ){
			res->nscount--;
			for(; nsx < res->nscount; nsx++ )
				res->NSLIST(nsx) = res->NSLIST(nsx+1);
		}
		return;
	}

	saddr = inet_ntoaV4(sin.sin_addr);

	if( strcmp(domain,RES_NSDOM0) == 0 ){
	  if( 0 <= nsx ){
		sin.sin_family = AF_INET;
		sin.sin_port = htons(PORT_DNS);
		res->NSLIST(nsx) = sin;
		debug(DBG_ANY,"        RES_NS[%d]=%s/%s\n",nsx,saddr,domain);
	  }else	debug(DBG_ANY," ignore RES_NS(%d)=%s/%s\n",MAXNS,ns,domain);
	}else{
		debug(DBG_ANY,"        RES_NS[%d]=%s/%s\n",nsx,saddr,domain);
		set_nameserver(domain,saddr);
	}
}
DNS_connect(addr,port)
	char *addr;
{	int sock,rcode;
	struct sockaddr_in sin;

	sock = socket(AF_INET,SOCK_STREAM,0);
	bzero(&sin,sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = inet_addrV4(addr);
	sin.sin_port = htons(port);
	if( connect(sock,(struct sockaddr*)&sin,sizeof(sin)) == 0 )
		return sock;

	close(sock);
	return -1;
}

RES_getns1(nsi,sin)
	SIN *sin;
{	State *res;

	res = &_res;
	if( nsi < _res.nscount ){
		*sin = res->NSLIST(nsi);
		return 1;
	}
	return 0;
}

load_rslvconf(res,path,loadns)
	State *res;
	char *path;
{	FILE *fp;
	char line[1024],com[1024],arg[1024];
	char *dp;
	char s[8][128];
	int sc,si;

	debug(DBG_ANY,"load_rslvconf(%s)\n",path);
	if( strncmp(path,"sh:",3) == 0 )
		fp = popen(path+3,"r");
	else	fp = fopen(path,"r");
	if( fp == NULL )
		return;

	while( fgets(line,sizeof(line),fp) != NULL ){
	    if( dp = strpbrk(line,"#;") )
		*dp = 0;
	    if( sscanf(line,"%s %s",com,arg) < 2 )
		continue;

	    if( strcmp(com,"debug") == 0 ){
		res->options |=  RES_DEBUG;
	    }else
	    if( strcmp(com,"nameserver") == 0 ){
		if( loadns ){
		    RES_ns1(res,arg,RES_NSDOM0);
		}
	    }else
	    if( strcmp(com,"domain") == 0 ){
		strcpy(res->defdname,arg);
		res->options |= RES_DEFNAMES;
	    }else
	    if( strcmp(com,"ndots") == 0 ){
		MIN_ABSNDOTS = atoi(arg);
	    }else
	    if( strcmp(com,"search") == 0 ){
		sc = sscanf(line,"%*s %s %s %s %s %s %s %s %s",
		    s[0],s[1],s[2],s[3],s[4],s[5],s[6],s[7]);

		if( loadns && 0 < sc ){
		    strcpy(res->defdname,s[0]);
		    for( si = 0; si < MAXDNSRCH && si < sc; si++ ){
			debug(DBG_ANY,"        RES_SRCH[%d] %s\n",si,s[si]);
			res->dnsrch[si] = strdup(s[si]);
		    }
		    res->dnsrch[sc] = 0;
		    res->options |=  RES_DNSRCH;
		    res->options &= ~RES_DEFNAMES;
		}
	    }else
	    if( strcmp(com,"sortlist") == 0 ){
		RES_scan_sortlist(line);
	    }
	}
	if( strncmp(path,"sh:",3) == 0 )
		pclose(fp);
	else	fclose(fp);
}

static ns1(name)
	char *name;
{	char nameb[128];
	char *domain;

	strcpy(nameb,name);
	if( domain = strchr(nameb,'/') )
		*domain++ = 0;
	else	domain = RES_NSDOM0;

	RES_ns1(&_res,nameb,domain);
	return 0;
}
RES_ns(nslist)
	char *nslist;
{
	if( nslist )
		scan_commaList(nslist,0,ns1);
	return 0;
}
RES_verify(verify)
	char *verify;
{
	RES_VERIFY = strdup(verify);
}
RES_verifyFaiure(host,badhost)
	char *host,*badhost;
{	char *sp,sc,*dp;

	dp = badhost;
	if( sp = RES_VERIFY ){
		for( ; sc = *sp; sp++ ){
			if( sc == '*' ){
				strcpy(dp,host);
				dp += strlen(dp);
			}else{
				*dp++ = sc;
			}
		}
	}
	*dp = 0;
}

RES_debug(debug)
	char *debug;
{
	_res.options |= RES_DEBUG;
	DNS_debug |= atoi(debug);
	return 0;
}
RES_domain(domain)
	char *domain;
{
	strcpy(_res.defdname,domain);
	return 0;
}
RES_order(order,porder)
	char *order,*porder;
{	int oi;
	char buff[RESOLVERS_SIZ];

	if( porder == NULL )
		porder = buff;
	strcpy(porder,resolvers);
	Strncpy(resolvers,order,sizeof(resolvers));

	debug(DBG_ANY,"RES_order(%s,%s)\n",resolvers,porder);
	return 0;
}
RES_conf(path)
	char *path;
{	static char *confpath;

	if( _RSLV_CONF == confpath )
		free(confpath);
	_RSLV_CONF = confpath = strdup(path);
	return 0;
}
RES_hosts(path)
	char *path;
{
	_HOSTSFILE = strdup(path);
	return 0;
}

RES_add_sortlist(addrs,masks,len)
	char *addrs[],*masks[];
{	int sx;

	for( sx = 0; addrs[sx]; sx++)
		RES_add_sortlist1(addrs[sx],masks?masks[sx]:NULL,len);
}
RES_add_sortlist1(addr,mask,len)
	unsigned char *addr,*mask;
{	AddrMask *sp;
	int ai,nm;
	unsigned char dmask[8];

	if( MAX_SORTLIST <= Nsortlist+1 ){
		debug(DBG_ANY,"!! exceed MAX_SORTLIST[%d]\n",MAX_SORTLIST);
		return;
	}
	if( mask == NULL || mask[0] == 0 ){
		mask = dmask;
		switch( ((int)addr[0] & 0xC0) >> 6 ){
		    case 0:  nm = 1; break;
		    case 2:  nm = 2; break;
		    default: nm = 3; break;
		}
		for( ai = 0; ai < nm ; ai++ )
			mask[ai] = 255;
		for( ; ai < len; ai++ )
			mask[ai] = 0;
	}
	debug(DBG_ANY,"        sortlist[%d] %d.%d.%d.%d / %d.%d.%d.%d\n",
		Nsortlist,
		addr[0],addr[1],addr[2],addr[3],
		mask[0],mask[1],mask[2],mask[3]);

	sp = &sort_list[Nsortlist++];
	for(ai = 0; ai < len; ai++){
		sp->addr[ai] = addr[ai];
		sp->mask[ai] = mask[ai];
	}
}
RES_scan_sortlist(line)
	char *line;
{	char s[16][128];
	char addr[32],mask[32];
	int a[16],m[16];
	int sc,si,ai;
	char ba[16],bm[16];
	int len = 4;

	sc = sscanf(line,"%*s %s %s %s %s %s %s %s %s %s %s",
	    s[0],s[1],s[2],s[3],s[4],s[5],s[6],s[7],s[8],s[9]);

	for( si = 0; si < sc; si++ ){
		addr[0] = mask[0] = 0;
		sscanf(s[si],"%[^/]/%s",addr,mask);
		for( ai = 0; ai < len; ai++ )
			a[ai] = m[ai] = 0;
		sscanf(addr,"%d.%d.%d.%d",&a[0],&a[1],&a[2],&a[3]);
		sscanf(mask,"%d.%d.%d.%d",&m[0],&m[1],&m[2],&m[3]);
		for( ai = 0; ai < len; ai++ ){
			ba[ai] = a[ai];
			bm[ai] = m[ai];
		}
		RES_add_sortlist1(ba,bm,len);
	}
}

sort_ipaddrs(addrs)
	unsigned char *addrs[];
{	unsigned char *ap,*mp;
	int si,sx,hit;

	hit = 0;
	for( si = 0; si < Nsortlist; si++ ){
		ap = sort_list[si].addr;
		mp = sort_list[si].mask;
		hit += sort_ipaddrs1(&addrs[hit],ap,mp);
	}
}

#define btohl(ap) ((ap[0]<<24) | (ap[1]<<16) | (ap[2]<<8) | ap[3])
sort_ipaddrs1(addrs,ap,mp)
	unsigned char *addrs[];
	unsigned char *ap,*mp;
{	unsigned char *cap,*saddrs[256],done[1];
	int ciaddr,cimasked;
	int sn,sx,dn,hit,imask;

	ciaddr = btohl(ap);
	imask = btohl(mp);
	cimasked = ciaddr & imask;

	for( sn = 0; addrs[sn]; sn++ )
		;

	dn = 0;
	for( sx = 0; cap = addrs[sx]; sx++ ){
		if( cap != done )
		if( (btohl(cap) & imask) == cimasked ){
			saddrs[dn++] = cap;
			addrs[sx] = done;
		}
	}
	hit = dn;
	for( sx = 0; sx < sn; sx++ )
		if( addrs[sx] != done )
			saddrs[dn++] = addrs[sx];
	for( sx = 0; sx < sn; sx++ )
		addrs[sx] = saddrs[sx];

	debug(DBG_ANY,"sort_ipaddrs(%d.%d.%d.%d/%d.%d.%d.%d) - %d/%d\n",
		ap[0],ap[1],ap[2],ap[3],mp[0],mp[1],mp[2],mp[3],hit,sn);
	return hit;
}
static res_getoptions(options,soptions)
	char *soptions;
{
	*soptions = 0;
	if( options & RES_DEBUG    ) strcat(soptions,"DEBUG ");
	if( options & RES_RECURSE  ) strcat(soptions,"RECURSE ");
	if( options & RES_DEFNAMES ) strcat(soptions,"DEFNAMES ");
	if( options & RES_DNSRCH   ) strcat(soptions,"DNSRCH ");
}
res_init()
{	char *env,*conf;
	char myname[128];
	char savorder[RESOLVERS_SIZ];
	int loadns;
	struct hostent *hp;
	char options[1024];
	int rx;
	char res1[RESOLVERS_SIZ],arg[RESOLVERS_SIZ];

	if( _res.options & RES_INIT )
		return;
	_res.options |= RES_INIT;

	if( env = getenv("RES_DEBUG") )
		RES_debug(env);

	if( _res.options & RES_DEBUG ){
		if( DNS_debug == 0 )
			DNS_debug = DBG_NS;
	}
	debug(DBG_ANY,"res_init()\n");
	Strncpy(savorder,resolvers,sizeof(savorder));
	strcpy(resolvers,"FND");

	Strncpy(resolvers,savorder,sizeof(resolvers));
	if( env = getenv("RES_ORDER") )
		Strncpy(resolvers,env,sizeof(resolvers));
	debug(DBG_ANY,"        RES_ORDER=%s\n",resolvers);

	for( rx = 0; rx = RES_next_res(resolvers,rx,res1,arg); ){
		if( res1[0] == 'D' && arg[0] != 0 )
			RES_ns1(&_res,arg,RES_NSDOM0);
	}

	if( env = getenv("RES_NS") )
		RES_ns(env);
	if( env = getenv("RES_VRFY") )
		RES_verify(env);

/*
	loadns = _res.nscount == 0;
*/
	loadns = 1; /* NSLIST is necessary for resolution of other NS ... */
	if( (conf = getenv("RES_CONF")) == 0 )
		conf = _RSLV_CONF;
	load_rslvconf(&_res,conf,loadns);

	if( env = getenv("RES_HOSTS") )
		RES_hosts(env);

	if( Nsortlist == 0 )
	if( gethostname(myname,sizeof(myname)) == 0 ){
	    if( hp = gethostbyname(myname) )
		RES_add_sortlist(hp->h_addr_list,NULL,hp->h_length);
	}

	if( _res.nscount == 0 )
	if( gethostname(myname,sizeof(myname)) == 0 ){
		unsigned char *ap;
		char saddr[32];
		int dnsock;

		if( hp = gethostbyname(myname) ){
			ap = (unsigned char*)hp->h_addr_list[0];
			/*ap[3] = 0xFF;  (broad cast in the segment) */
			sprintf(saddr,"%d.%d.%d.%d",ap[0],ap[1],ap[2],ap[3]);
		}else	strcpy(saddr,"127.0.0.1");

		if( 0 <= (dnsock = DNS_connect(saddr,PORT_DNS)) ){
			close(dnsock);
			RES_ns1(&_res,saddr,RES_NSDOM0);
		}
	}
	if( (env = getenv("RES_DOMAIN")) || (env = getenv("LOCALDOMAIN")) ){
		strcpy(_res.defdname,env);
		_res.options &= ~RES_DNSRCH;
		_res.options |=  RES_DEFNAMES;
	}

	if( _res.defdname[0] )
		debug(DBG_ANY,"        RES_DOMAIN=%s\n",_res.defdname);

	res_getoptions(_res.options,options);
	debug(DBG_ANY,"        options = %s\n",options);
}

char resolv_errmsg[512];
int (*RES_debugprinter)();
res_debug(flag,fmt,a,b,c,d,e,f,g,h,i,j,k,l,m,n)
	char *fmt,*a,*b,*c,*d,*e,*f,*g,*h,*i,*j,*k,*l,*m,*n;
{	int now;

	if( flag == DBG_FORCE || _res.options & RES_DEBUG && DNS_debug & flag ){
		now = time(0);
		if( RES_debugprinter )
			(*RES_debugprinter)(fmt,a,b,c,d,e,f,g,h,i,j,k,l,m,n);
		else{
			fprintf(stderr,"%02d:%02d ",(now%3600)/60,now%60);
			fprintf(stderr,fmt,a,b,c,d,e,f,g,h,i,j,k,l,m,n);
		}
		return 1;
	}else	return 0;
}
int (*RES_log)();
res_log(which,byname,name,rv,cname)
	char *name,*rv[],*cname;
{
	if( RES_log )
		(*RES_log)(which,byname,name,rv,cname);
}

static putNames(hp)
	struct hostent *hp;
{	unsigned char *op;
	int hi;

	printf("%s",hp->h_name);
	for( hi = 0; ;hi++){
		op = (unsigned char*)hp->h_aliases[hi];
		if( op == NULL )
			break;
		printf(",%s",op);
	}
}
static putAddrs(hp)
	struct hostent *hp;
{	unsigned char *op;
	int hi;

	for( hi = 0; ; hi++ ){
		op = (unsigned char*)hp->h_addr_list[hi];
		if( op == NULL )
			break;
		if( 0 < hi )
		printf(",");
		printf("%d.%d.%d.%d",op[0],op[1],op[2],op[3]);
	}
}

struct hostent *RES_gethost(addrhost)
	char *addrhost;
{	struct hostent *hp;
	int ia[4],ii;
	char baddr[4];

	if( inet_addrV4(addrhost) == -1 )
		hp = gethostbyname(addrhost);
	else{
		sscanf(addrhost,"%d.%d.%d.%d",&ia[0],&ia[1],&ia[2],&ia[3]);
		for( ii = 0; ii < 4; ii++ )
			baddr[ii] = ia[ii];
		hp = gethostbyaddr(baddr,sizeof(baddr),AF_INET);
	}
	return hp;
}
RES_1(f_aton,fp,arg)
	FILE *fp;
	char *arg;
{	struct hostent *hp;

	if( f_aton ){
		if( inet_addrV4(arg) != -1 ) /* is IP address */
			if( hp = RES_gethost(arg) )
				arg = hp->h_name;
		printf("%s\n",arg);
		return;
	}
	hp = RES_gethost(arg);

	if( hp ){
		unsigned char *op;
		int hi;
		putAddrs(hp);
		printf("\t");
		putNames(hp);
		printf("\n");
	}else{
		if( inet_addrV4(arg) == -1 )
			printf("?\t%s\n",arg);
		else	printf("%s\t?\n",arg);
		/*exit(1);*/
	}
}
RES_1s(addrhost,addr_host)
	char *addrhost;
	char *addr_host;
{	struct hostent *hp;
	unsigned char *op,addr[32];

	if( hp = RES_gethost(addrhost) ){
		op = (unsigned char *)hp->h_addr;
		sprintf(addr,"%d.%d.%d.%d",op[0],op[1],op[2],op[3]);
		sprintf(addr_host,"%s\t%s\n",addr,hp->h_name);
		return 1;
	}
	return 0;
}

extern int RSLV_TIMEOUT;
RES_timeout(timeout)
{
	RSLV_TIMEOUT = timeout;
	return 1;
}

