/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1997 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	filter.c (external/internal filter for each connection)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:

    remote filter ...

        FFROMCL=cfi://host:port
	Filter: cfi://host:port
	CGI:    cgi://host:port

History:
	970814	extracted from service.c and delegated.c
//////////////////////////////////////////////////////////////////////#*/
#include "delegate.h"
#include "vsignal.h"
#include "param.h"

#define FTOCL_FIELD     "X-Request"

extern char **environ;
extern char *DELEGATE_getEnv();
extern char *strfConn();
extern char *malloc();
extern FILE *URLget();
extern char D_RPORT[];

#define F_CL		 1
#define F_TOCL		 2
#define F_FROMCL	 3
#define F_SV		 4
#define F_TOSV		 5
#define F_FROMSV	 6
#define F_MD		 7
#define F_TOMD		 8
#define F_FROMMD	 9
#define F_RPORT		10

#define F_CACHE		 0
#define F_TOCACHE	 0
#define F_FROMCACHE	 0
#define F_PROXY		 0

typedef struct {
	char	*f_filter;
} Filter;
static Filter filters[16];

#define FS_FCL		filters[F_CL].f_filter
#define FS_FFROMCL	filters[F_FROMCL].f_filter
#define FS_RPORT	filters[F_RPORT].f_filter
#define FS_FTOCL	filters[F_TOCL].f_filter
#define FS_FTOSV	filters[F_TOSV].f_filter
#define FS_FFROMSV	filters[F_FROMSV].f_filter
#define FS_FFROMMD	filters[F_FROMMD].f_filter
#define FS_FTOMD	filters[F_TOMD].f_filter

#define CFIMAGIC	"#!cfi"
#define isCFI(filter)	(strncmp(filter,CFIMAGIC,strlen(CFIMAGIC)) == 0)

filter_isCFI(which)
{	char *filter;

	if( which == XF_FTOCL )
		return FS_FTOCL != NULL && isCFI(FS_FTOCL);
	if( which == XF_FTOSV )
		return FS_FTOSV != NULL && isCFI(FS_FTOSV);
	return 0;
}

static sigTERM(sig)
{
	exit(0);
}

static char *open_cfi(file)
	char *file;
{	FILE *fp;
	char desc[128];
	char convspec[0x10000];
	char *sp;
	int size;
	int off;

	fp = fopen(file,"r");
	if( fp == NULL )
		if( isFileURL(file) )
			fp = URLget(file,1,NULL);

	sp = NULL;
	if( fp != NULL ){
		desc[0] = 0;
		off = ftell(fp);
		fgets(desc,sizeof(desc),fp);
		fseek(fp,off,0);

		if( strncmp(desc,CFIMAGIC,strlen(CFIMAGIC)) == 0 ){
			size = file_size(fileno(fp));
			sp = malloc(size+1);
			fread(sp,1,size,fp);
			sp[size] = 0;
		}
		fclose(fp);
	}
	return sp;
}

static setF1(fi,filter)
	char *filter;
{	char *cfi;

	if( filter && filters[fi].f_filter == NULL ){
		if( cfi = open_cfi(filter) )
			filters[fi].f_filter = cfi;
		else	filters[fi].f_filter = strdup(filter);
	}
}

char *getFTOCL(){ return FS_FTOCL; }
setFTOCL(ftocl) char *ftocl; { FS_FTOCL = ftocl; }
scan_FTOCL(filter) char *filter; { setF1(F_TOCL,filter); }

char *getFTOSV(){ return FS_FTOSV; }
setFTOSV(ftosv) char *ftosv; { FS_FTOSV = ftosv; }
scan_FTOSV(filter) char *filter; { setF1(F_TOSV,filter); }


scan_FILTERS(Conn)
	Connection *Conn;
{
	setF1(F_CL,     DELEGATE_getEnv(P_FCL));
	setF1(F_FROMCL, DELEGATE_getEnv(P_FFROMCL));
	setF1(F_TOCL,   DELEGATE_getEnv(P_FTOCL));

/*	setF1(F_SV,     DELEGATE_getEnv(P_FSV)); */
	setF1(F_FROMSV, DELEGATE_getEnv(P_FFROMSV));
	setF1(F_TOSV,   DELEGATE_getEnv(P_FTOSV));

/*	setF1(F_MD,     DELEGATE_getEnv(P_MD)); */
	setF1(F_TOMD,   DELEGATE_getEnv(P_FTOMD));
	setF1(F_FROMMD, DELEGATE_getEnv(P_FFROMMD));

	FS_RPORT    = DELEGATE_getEnv(P_RPORT);
      /*scan_RPORT(Conn,DELEGATE_getEnv(P_RPORT));*/
}


getTimestamp(stime)
	char *stime;
{	int now,usec;

	now = Gettimeofday(&usec);
	StrftimeLocal(stime,sizeof(stime),"%m/%d-%H:%M:%S%.3s",now,usec);
}
teeThru(in,outfp,teefp)
	FILE *in,*outfp,*teefp;
{	int ch;

	while( 0 < ready_cc(in) ){
		ch = getc(in);
		if( ch == EOF )
			return;
		if( outfp ) putc(ch,outfp);
		if( teefp ) putc(ch,teefp);
	}
	if( outfp ) fflush(outfp);
	if( teefp ) fflush(teefp);

	if( outfp == 0 ) simple_relay(fileno(in),fileno(teefp)); else
	if( teefp == 0 ) simple_relay(fileno(in),fileno(outfp)); else
			 relay_tee(fileno(in),fileno(outfp),fileno(teefp));
}

#define STAT	1
#define HEAD	2
#define	EOH	3
#define BODY	4
#define	EOR	5
#define BINARY	6

/*   TEEFILTER
 *	Output to "outfp" is through always to a client/servers
 *	Output to "teefp" may be prefixed with something by "-t -p -n"
 *	When in "tee" mode with "-h -b", only specified part of input will
 *	be relayed to "teefp"
 *	When in non-"tee" mode (maybe in "cat" mode), "teefp" is directed
 *	to original "outfp".  All of output is relayed to "teefp" in this
 *	case and "-h -b" controls to which parts "-t -p -n" is applied
 *	selectively.
 */
teeFilter(infp,outfp,filter,opts,tee)
	FILE *infp,*outfp;
	char *filter,*opts;
{	FILE *teefp;
	char line[1024];
	char *dp;
	int headonly = 0;
	int bodyonly = 0;
	int append = 0;
	int withlnum = 0;
	int processid = 0;
	int timestamp = 0;
	int nl = 0,nb = 0,NL = 0;
	int where;
	int pid;
	char stime[32];

	dp = opts;
	where = STAT;

	for(;;){
		if( strncmp(dp,"-s",2)==0 ){ where = HEAD;  dp += 2; }else
		if( strncmp(dp,"-h",2)==0 ){ headonly = 1;  dp += 2; }else
		if( strncmp(dp,"-b",2)==0 ){ bodyonly = 1;  dp += 2; }else
		if( strncmp(dp,"-t",2)==0 ){ timestamp = 1; dp += 2; }else
		if( strncmp(dp,"-p",2)==0 ){ processid = 1; dp += 2; pid = getpid(); }else
		if( strncmp(dp,"-n",2)==0 ){ withlnum = 1;  dp += 2; }else
		if( strncmp(dp,"-a",2)==0 ){ append = 1;    dp += 2; }else
			break;
	}
	
	if( dp[0] == 0 )
		teefp = stderr;
	else
	if( append )
		teefp = fopen(dp,"a");
	else	teefp = fopen(dp,"w");
	if( teefp == NULL ){
		sv1log("#### -tee: cannot open %s\n",dp);
		return 0;
	}
	if( !tee ){
		teefp = outfp;
		outfp = NULL;
	}

	for(;;){
		if( fPollIn(infp,100) == 0 ){
			if( outfp ) fflush(outfp);
			if( teefp ) fflush(teefp);
		}
		if( fgets(line,sizeof(line),infp) == NULL )
			break;
		nb += strlen(line);
		nl++;

		if( where <= HEAD && strncmp(line,FTOCL_FIELD,strlen(FTOCL_FIELD))==0 )
			continue;

		if( line[0] == '.' && (line[1] == '\n' || line[1] == '\r') )
			where = BINARY;

		if( outfp )
			fputs(line,outfp);

		if( teefp ){
			if( !tee && where == STAT )
				fputs(line,teefp);
			else
			if( headonly && where != HEAD || bodyonly && where != BODY ){
				if( !tee )
				fputs(line,teefp);
			}else{
				NL++;
				if( timestamp ){
					getTimestamp(stime);
					fprintf(teefp,"%s ",stime);
				}
				if( processid )
					fprintf(teefp,"[%d] ",pid);
				if( withlnum )
					fprintf(teefp,"%6d\t",NL);
				fputs(line,teefp);
			}
		}

		if( where == STAT )
			where = HEAD;

		if( where <= HEAD  && (line[0] == '\n' || line[0] == '\r') )
			where = EOH;

		if( where == EOH ){
			where = BODY;
			if( headonly ){
				if( tee )
					simple_relayf(infp,outfp);
				else	simple_relayf(infp,teefp);
				break;
			}
		}
	}
	if( tee && teefp != stderr )
		fclose(teefp);

	Verbose("#### %s [%d] bytes / [%d] lines\n",filter,nb,nl);
}

remote_cfi(filter,in,out)
	char *filter;
	FILE *in,*out;
{	char host[1024],options[1024];
	int port,sock;
	int sv[2][2],rv[2],xv[2];
	FILE *ts,*fs;
	int ch;

	host[0] = 0;
	port = 0;
	options[0] = 0;
	if( sscanf(filter,"cfi://%[^:]:%d/%s",host,&port,options) ){
		sock = client_open("cfi","data",host,port);
		if( sock < 0 )
			return -1;
	}else	return -1;

	ts = fdopen(sock,"w");
	if( options[0] ){
		fprintf(ts,"POST /%s HTTP/1.0\r\n",options);
		fprintf(ts,"\r\n");
		fflush(ts);
	}
	while( 0 < ready_cc(in) ){
		ch = getc(in);
		putc(ch,ts);
	}
	fflush(ts);

	sv[0][0] = fileno(in); sv[0][1] = sock;        xv[0] = 1;
	sv[1][0] = sock;       sv[1][1] = fileno(out); xv[1] = 0;
	relaysx(0,2,sv,xv,rv,NULL,NULL);

	fclose(ts);
	return 0;
}

extern char *wordscan();
builtin_filter(filter,in,out)
	char *filter;
	FILE *in,*out;
{	int rcc;

	if( strncmp(filter,"cfi:",4) == 0 ){
		remote_cfi(filter,in,out);
		return 1;
	}
	if( strcmp(filter,"-thru") == 0 ){
		rcc = simple_relayf(in,out);
		Verbose("#### %s [%d] bytes\n",filter,rcc);
		return 1;
	}
	if( strncmp(filter,"-cat",4) == 0 ){
		teeFilter(in,out,filter,filter+4,0);
		return 1;
	}else
	if( strncmp(filter,"-tee",4) == 0 ){
		teeFilter(in,out,filter,filter+4,1);
		return 1;
	}
	if( strcmp(filter,"-jis") == 0 ){
		rcc = CCV_relay_textX("JIS.JP",in,out);
		Verbose("#### %s [%d]\n",filter,rcc);
		return 1;
	}
	if( strcmp(filter,"-euc") == 0 ){
		rcc = CCV_relay_textX("EUC.JP",in,out);
		Verbose("#### %s [%d]\n",filter,rcc);
		return 1;
	}
	if( strcmp(filter,"-sjis") == 0 ){
		rcc = CCV_relay_textX("SJIS.JP",in,out);
		Verbose("#### %s [%d]\n",filter,rcc);
		return 1;
	}
	return -1;
}

static char **addConnEnviron(Conn,ev,ebuff)
	Connection *Conn;
	char *ev[];
	char *ebuff;
{	char **nev,*ep;
	int ei;

	for( ei = 0; ev[ei]; ei++ )
		;
	nev = (char**)malloc((ei+5)*sizeof(char*));
	for( ei = 0; ev[ei]; ei++ )
		nev[ei] = ev[ei];
	ep = ebuff;
	nev[ei++] = ep; ep = strfConn(Conn,"REMOTE_IDENT=%u",ep) + 1;
	nev[ei++] = ep; ep = strfConn(Conn,"REMOTE_HOST=%h",ep) + 1;
	nev[ei++] = ep; ep = strfConn(Conn,"REMOTE_ADDR=%a",ep) + 1;
	*ep = 0;
	nev[ei++] = 0;
	return nev;
}
static execFilter(Conn,in,out,ifd,ofd,what,isresp,filter)
	Connection *Conn;
	FILE *in,*out;
	char *what,*filter;
{	int fdx;
	char *av[128],ac,ai;
	char abuff[0x2000];
	char ebuff[1024];
	int tofil[2];

	/* this may be harmful for Windows
	 * because it may close client socket ... */
	/* if( INHERENT_fork() ){
		int logfd;
		logfd = curLogFd();
		for( fdx = 3; fdx < 32; fdx++ )
			if( fdx != logfd )
				close(fdx);
	}
	 */

	if( in  == NULL ) in = fdopen(ifd,"r");
	if( out == NULL ) out = fdopen(ofd,"w");

	if( 0 <= builtin_filter(filter,in,out) ){
		fflush(out);
		Finish(0);
	}


	if( isCFI(filter) ){
		char conninfo[2048];

		sv1log("#### execFilter[%s] CFI\n",what);
		make_conninfo(Conn,conninfo,filter);

		environ = addConnEnviron(Conn,environ,ebuff);
		cfi(isresp,in,out,conninfo,filter);
		Finish(0);
	}else{
		sv1log("#### execFilter[%s] %s\n",what,filter);
		ac = 0;
		av[ac++] = "sh";
		av[ac++] = "-c";
		av[ac++] = filter;
		ac = DELEGATE_copyEnv(av,ac,filter,abuff);
		av[ac] = 0;

		environ = addConnEnviron(Conn,environ,ebuff);

		if( 0 < ready_cc(in) ){
			sv1log("#### relay buffered input\n");
			pipe(tofil);
			if( Fork("FILTER-INBUFF") == 0 ){
				FILE *tout;
				close(tofil[0]);
				tout = fdopen(tofil[1],"w");
				simple_relayf(in,tout);
				Finish(0);
			}else{
				close(tofil[1]);
				ifd = tofil[0];
			}
		}
		dup2(ifd,0);
		dup2(ofd,1);
		Execvp(what,"sh",av);
		Finish(-1);
	}
}
static callF1(Conn,in,out,filter)
	Connection *Conn;
	FILE *in,*out;
	char *filter;
{
	execFilter(Conn,in,out,fileno(in),fileno(out),
		Conn->fi_what,Conn->fi_isresp,filter);
}
static forkexecFilter1X(Conn,dst,src,what,iomode,isresp,filter,pidp)
	Connection *Conn;
	char *what,*filter;
	int *pidp;
{	int tofil[2],fd;
	int pid;

	pipe(tofil);
	if( INHERENT_fork() ){
		if( (pid = Fork(what)) == 0 ){
			ProcTitle(Conn,"(filter:%s)",what);
			signal(SIGTERM,sigTERM);
			signal(SIGINT, sigTERM);
			if( iomode & 1 ){
				if( 0 <= src ) close(src);
				close(tofil[1]);
				execFilter(Conn,NULL,NULL,tofil[0],dst,
					what,isresp,filter);
			}else{
				if( 0 <= dst ) close(dst);
				close(tofil[0]);
				execFilter(Conn,NULL,NULL,src,tofil[1],
					what,isresp,filter);
			}
		}
		if( iomode & 1 ){
			close(tofil[0]);
			if( pidp ) *pidp = pid;
			return tofil[1];
		}else{
			close(tofil[1]);
			if( pidp ) *pidp = pid;
			return tofil[0];
		}
	}else{
		strcpy(Conn->fi_what,what);
		Conn->fi_isresp = isresp;

		if( iomode & 1 ){
			pid = spawnFilter(Conn,iomode,tofil,dst,callF1,filter);
			close(tofil[0]);
			if( pidp ) *pidp = pid;
			return tofil[1];
		}else{
			pid = spawnFilter(Conn,iomode,tofil,src,callF1,filter);
			close(tofil[1]);
			if( pidp ) *pidp = pid;
			return tofil[0];
		}
	}
}

#define IS_BUILTIN(filter)	(filter[0] == '-')

static forkexecFilter1(Conn,dst,src,what,iomode,isresp,filter,pidp)
	Connection *Conn;
	char *what,*filter;
	int *pidp;
{	char xwhat[64];
	int sock,psock,fsock,pid;

	if( pidp != NULL )
		*pidp = 0;

	if( iomode & 1 )
		sock = dst;
	else	sock = src;

	if( pidp == NULL )
	if( IS_BUILTIN(filter)
	 || WithSocketFile() || !file_ISSOCK(sock)
	)
	return forkexecFilter1X(Conn,dst,src,what,iomode,isresp,filter,NULL);

	if( !isresp && DFLT_PROTO[0] == 0 && isCFI(filter) ){
		extern int IO_TIMEOUT;
		int rcc;
		char buff[32];

		if( 0 < PollIn(src,IO_TIMEOUT*1000) ){
			setNonblockingIO(src,1);
			rcc = RecvPeek(src,buff,16);
			setNonblockingIO(src,0);
			if( 0 < rcc ){
				buff[rcc] = 0;
				if( HTTP_isMethod(buff) )
					strcpy(DFLT_PROTO,"http");
			}
		}
	}

	sprintf(xwhat,"%s-P",what);
	psock = forkexecFilter1X(Conn,dst,src,xwhat,iomode,isresp,"-thru",&pid);
	if( pidp != NULL )
		*pidp = pid;
	sv1log("#### pre-filter inserted: %d\n",pid);

	if( iomode & 1 )
		dst = psock;
	else	src = psock;

	fsock = forkexecFilter1X(Conn,dst,src,what, iomode,isresp,filter, NULL);
	close(psock);
	return fsock;
}

/*
 * RPORT : redirection port for response from the MASTER
 */
scan_RPORT(Conn,portin)
	Connection *Conn;
	char *portin;
{	char host[256],tcp_udp[128];
	int sock,port,listen;

	RPORTsock = -1;
	if( portin == NULL )
		return;

	tcp_udp[0] = host[0] = 0;
	port = 0;
	sscanf(portin,"%[^:]:%[^:]:%d",tcp_udp,host,&port);
	if( strcasecmp(tcp_udp,"tcp") == 0 ) listen =  1; else
	if( strcasecmp(tcp_udp,"udp") == 0 ) listen = -1; else{
		sv1tlog("%s ? %s\n",tcp_udp,portin);
		Finish(-1);
	}
	RPORTudp = (listen < 0);
	sock = server_open("RPORT",host,port,listen);
	sockHostport(sock,&port);
	if( host[0] == 0 )
		gethostname(host,sizeof(host));

	sprintf(D_RPORT,"%s:%s:%d",tcp_udp,host,port);
	RPORTsock = sock;
}
static accept_RPORT(Conn)
	Connection *Conn;
{	int sock;

	if( RPORTsock < 0 )
		return -1;

	sv1log("ACCEPT RPORT[%d]...\n",RPORTsock);
	if( RPORTudp ){
		sock = RPORTsock;
		sv1log("ACCEPT RPORT[%d][%d] UDP\n",RPORTsock,sock);
	}else{
		sock = ACCEPT(RPORTsock,0,-1,0);
		sv1log("ACCEPT RPORT[%d][%d]\n",RPORTsock,sock);
		close(RPORTsock);
	}
	RPORTsock = -1;
	return sock;
}
static connect_RPORTX(Conn,portin)
	Connection *Conn;
	char *portin;
{	char host[256],tcp_udp[128];
	int sock,port,listen;

	if( D_RPORTX[0] == 0 )
		return -1;

	sv1log("CONNECT RPORTX[%s]\n",portin);
	if( sscanf(portin,"%[^:]:%[^:]:%d",tcp_udp,host,&port) != 3 )
		return -1;

	if( strcasecmp(tcp_udp,"udp") == 0 )
		sock = UDP_client_open("RPORTX","raw",host,port);
	else	sock = client_open("RPORTX","raw",host,port);
	return sock;
}
insertFTOCL(Conn,client,server)
	Connection *Conn;
{	char *filter;
	int fsock;

	if( FS_FTOCL == NULL )
		return client;
	filter = FS_FTOCL;

	fsock = forkexecFilter1(Conn,client,server,"FTOCL",  1,1,filter,NULL);

	EchoRequest = 1;
	Conn->xf_filters |= XF_FTOCL;
	return fsock;
}
static insertFFROMCL(Conn,client,fpidp)
	Connection *Conn;
	int *fpidp;
{	char *filter;
	int fromclp[2];
	int fsock;

	if( FS_FFROMCL == NULL )
		return -1;
	filter = FS_FFROMCL;

	/* V.3.0.12 insert a simple relay process
	 * to cause normal EOF at the input of FFROMCL filter ?
	 */
	fsock = forkexecFilter1(Conn,-1,client,"FFROMCL", 0,0,filter,fpidp);
	Conn->xf_filters |= XF_FFROMCL;
	return fsock;
}
static insertFCL(Conn,fromC)
	Connection *Conn;
{	char *filter;
	int fromcl[2];

	if( FS_FCL == NULL )
		return -1;
	filter = FS_FCL;

	Socketpair(fromcl);
	if( Fork("FCL") == 0 ){
		close(fromcl[0]);
		execFilter(Conn,NULL,NULL,fromC,fromcl[1],"FCL",2,filter);
	}
	close(fromcl[1]);
	Conn->xf_filters |= XF_FCL;
	return fromcl[0];
}

insertFTOSV(Conn,client,server)
	Connection *Conn;
{	char *filter;
	int fsock;

	if( FS_FTOSV == NULL )
		return server;
	filter = FS_FTOSV;

	fsock = forkexecFilter1(Conn,server,client,"FTOSV",  1,0,filter,NULL);
	Conn->xf_filters |= XF_FTOSV;
	return fsock;
}
static insertFFROMSV(Conn,client,server)
	Connection *Conn;
{	char *filter;
	int fsock;

	if( FS_FFROMSV == NULL )
		return server;

	filter = FS_FFROMSV;

	fsock = forkexecFilter1(Conn,client,server,"FFROMSV",0,1,filter,NULL);
	Conn->xf_filters |= XF_FFROMSV;
	return fsock;
}

static insert2IN1(dst,src)
{	int pipe[2];
	int fdv[2][2],ccv[2];

	Socketpair(pipe);
	if( Fork("FFROMMD-1") == 0 ){
		/* DeleGate -> MASTER (through) */
		fdv[0][0] = pipe[1];
		fdv[0][1] = dst;

		/* MASTER -> filter -> DeleGate */
		fdv[1][0] = src;
		fdv[1][1] = pipe[1];

		close(pipe[0]);
		relays(0,2,fdv,ccv);
		Finish(0);
	}
	close(pipe[1]);

	close(src);
	return pipe[0];
}
static insertFFROMMD(Conn,master)
	Connection *Conn;
{	char *filter;
	int fsock;

	if( FS_FFROMMD == NULL )
		return -1;
	filter = FS_FFROMMD;

	fsock = forkexecFilter1(Conn,-1,    master,"FFROMMD",0,1,filter,NULL);
	fsock = insert2IN1(master,fsock);

	Conn->xf_filters |= XF_FFROMMD;
	return fsock;
}
static insertFTOMD(Conn,master)
	Connection *Conn;
{	int pipe[2];

	if( FS_FTOMD == NULL )
		return -1;

	return -1;
}


/*
 * Insert filters before the invocation of protocol interpreters
 * Filters which proces all of data from client (FCL and FFROMCL)
 * must be invoked at the start of interpretation.
 */
insert_FCLIENTS(Conn,fromCp,toCp)
	Connection *Conn;
	int *fromCp,*toCp;
{	char *filter;
	int fpid;
	int fromC,toC,toF;

	fromC = *fromCp;
	toC = *toCp;
	fpid = 0;

	if( 0 <= (toF = insertFCL(Conn,fromC)) )
		fromC = toC = toF;
	else
	if( 0 <= (toF = insertFFROMCL(Conn,fromC,&fpid)) )
		fromC = toF;

	scan_RPORT(Conn,FS_RPORT);

	*fromCp = fromC;
	*toCp = toC;
	return fpid;
}

/*
 * Insert filters right after the connection establishment to the server
 */
insert_FSERVER(Conn,fromC)
	Connection *Conn;
{	int toS;

	if( ImCC ) return;

	if( 0 <= ToS ){
		toS = ToS;
		ToS = insertFTOSV(Conn,fromC,ToS);

		if( ToS != toS )
			ToSX = toS;
		else	ToSX = -1;
	}

	if( 0 <= FromS )
		FromS = insertFFROMSV(Conn,fromC,FromS);
}
close_FSERVER(Conn,realclose)
	Connection *Conn;
{
	if( 0 <= ToSX ){
		close(ToS);
		if( realclose ){
			close(ToSX);
			ToS = -1;
		}else{
			ToS = ToSX;
			/* maybe ImCC */
		}
		ToSX = -1;

		if( Conn->xf_filters & XF_FTOSV ){
			Conn->xf_filters &= ~XF_FTOSV;
			if( (Conn->xf_isremote & XF_FTOSV) == 0 )
				wait(0);
			Conn->xf_isremote &= ~XF_FTOSV;
		}
	}
}

/*
 * Insert filters right after the connection establishment to the MASTER
 */
insert_FMASTER(Conn,msock)
	Connection *Conn;
{	int fromS,toS;

	if( (fromS = accept_RPORT(Conn)) < 0 )
		fromS = msock;

	if( (FromS = insertFFROMMD(Conn,fromS)) < 0 )
		FromS = fromS;
	else	FromSX = FromS; /* FFROMMD is inserted */

	if( 0 <= (toS = insertFTOMD(Conn,msock)) ){
		ToS = toS;
		/*ToSX = ToS;*/
	}
}

/*
 *  Insert filters after some protocol (HTTP) header interpretation
 */
insert_FPROTO(Conn,toCp,toS,fromSp)
	Connection *Conn;
	int *toCp,*fromSp;
{	int xtoC,fromS,toC;

	fromS = *fromSp;
	toC = *toCp;

	if( fromS < 0 )
		fromS = toS;

	if( 0 <= (xtoC = connect_RPORTX(Conn,D_RPORTX)) ){
		if( toC != Conn->cl_sock )
			close(toC);
		toC = xtoC;
	}
	if( 0 <= (xtoC = insertFTOCL(Conn,toC,toS)) && xtoC != toC ){
		if( toC != Conn->cl_sock )
			close(toC);
		toC = xtoC;
	}

	*fromSp = fromS;
	*toCp = toC;
}


filter_withCFI(Conn,which)
	Connection *Conn;
{
	if( ImCC == 0 )
	if( Conn->xf_filters & which )
	if( filter_isCFI(which) )
		return 1;
	return 0;
}

static callFsystem(Conn,in,out,command)
	Connection *Conn;
	FILE *in,*out;
	char *command;
{	int fd;

	dup2(fileno(in),0);
	dup2(fileno(out),1);
	dup2(fileno(out),2);
	for( fd = 3; fd < 32; fd++ )
		close(fd);
	system(command);
}

systemFilter(command,in,out)
	char *command;
	FILE *in,*out;
{	int tofd[2];
	int fd0,fd1;
	int code;
	FILE *tofp;
	char *av[4];
	int pid;
	Connection ConnBuf, *Conn = &ConnBuf;

	fflush(out);
	if( ready_cc(in) <= 0 ){
		sv1log("systemFilter: %s\n",command);
		fd0 = dup(0); dup2(fileno(in), 0);
		fd1 = dup(1); dup2(fileno(out),1);
		code = system(command);
		dup2(fd0,0); close(fd0);
		dup2(fd1,1); close(fd1);
		return 0;
	}

	sv1log("systemFilter[with buffered input = %d]: %s\n",
		ready_cc(in),command);

	pipe(tofd);
	if( INHERENT_fork() ){
		if( (pid = fork()) == 0 ){
			close(tofd[1]);
			close(fileno(in));
			dup2(tofd[0],0);
			dup2(fileno(out),1);
			system(command);
			exit(0);
		}
	}else{
		pid = spawnFilter(Conn,1,tofd,fileno(out),callFsystem,command);
	}
	close(tofd[0]);
	tofp = fdopen(tofd[1],"w");
	simple_relayf(in,tofp);
	fclose(tofp);

	wait(0);
	return pid;
}
