/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1997 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	resource.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	970813	extracted from delegated.c
//////////////////////////////////////////////////////////////////////#*/
#include "delegate.h"
#include "param.h"
extern FILE *TMPFILE();

extern int CII_getFonts();
extern int CCX_getTable();

static char *CCXserver;
static char *CIIserver;

static load1(Conn,url,charset,func,what)
	Connection *Conn;
	char *url,*charset;
	int (*func)();
	char *what;
{	FILE *tmp;
	int size,got;

	tmp = TMPFILE(what);
	URLget(url,0,tmp);
	size = file_size(fileno(tmp))-ftell(tmp);
	if( 0 < size )
		got = (*func)(charset,tmp);
	else	got = 0;
	fclose(tmp);
	sv1log("%s: %s: size=%d/%d\n",what,url,got,size);
	return 0;
}
static loadCIIfont1(Conn,url,charset)
	Connection *Conn;
	char *url,*charset;
{
	load1(Conn,url,charset,CII_getFonts,"loadCIIfonts");
	return 0;
}

static loadCIIfont0(Conn,url)
	Connection *Conn;
	char *url;
{	char fullurl[1024];

	if( !isFullURL(url) ){
		strcpy(fullurl,CIIserver);
		if( strtailchr(CIIserver) != '/' && url[0] != '/' )
			strcat(fullurl,"/");
		strcat(fullurl,url);
		url = fullurl;
	}
	return loadCIIfont1(Conn,url,"*");
}
static loadCCXtable1(Conn,url,charset)
	Connection *Conn;
	char *url,*charset;
{
	load1(Conn,url,charset,CCX_getTable,"loadCCXtable");
	return 0;
}
static loadCCXtable0(Conn,url,ccxserver)
	Connection *Conn;
	char *url;
	char *ccxserver;
{	char fullurl[1024];

	if( !isFullURL(url) ){
		strcpy(fullurl,ccxserver);
		if( strtailchr(ccxserver) != '/' && url[0] != '/' )
			strcat(fullurl,"/");
		strcat(fullurl,url);
		url = fullurl;
	}
	loadCCXtable1(Conn,url,"*");
}

static loadCIIfont2(Conn,charcode)
	Connection *Conn;
	char *charcode;
{	char charset[128],url[1024];

	if( need_CCXtable(charcode,charset) ){
		Verbose("need CCXtable of %s\n",charset);
		if( CCX_withTable(charset) )
			return 0;
		strcpy(url,CCXserver);
		if( strtailchr(url) != '/' )
			strcat(url,"/");
		strcat(url,charset);
		loadCCXtable1(Conn,url,charset);
	}
	if( need_CIIfont(charcode,charset) ){
		Verbose("need CIIfont of %s\n",charset);
		if( CII_withCharset(charset) )
			return 0;
		strcpy(url,CIIserver);
		if( strtailchr(url) != '/' )
			strcat(url,"/");
		strcat(url,charset);
		loadCIIfont1(Conn,url,charset);
	}
	return 0;
}

extern char *DELEGATE_CCXSERVER;
extern char *DELEGATE_CIISERVER;
extern char *DELEGATE_getEnv();

load_resources(Conn,ccxserver,ciiserver)
	Connection *Conn;
	char *ccxserver,*ciiserver;
{	char fonts[2048],table[2048];

	if( (CCXserver = DELEGATE_getEnv(P_CCXSERVER)) == NULL )
		CCXserver = DELEGATE_CCXSERVER;
	if( (CIIserver = DELEGATE_getEnv(P_CIISERVER)) == NULL )
		CIIserver = DELEGATE_CIISERVER;

	DELEGATE_scanEnv(Conn,P_CCXTABLE,loadCCXtable0);
	DELEGATE_scanEnv(Conn,P_CIIFONT, loadCIIfont0);
	DELEGATE_scanEnv(Conn,P_CHARCODE,loadCIIfont2);

	if( CCX_lsTable(table,sizeof(table)) )
		Verbose("CCXtable:\n%s",table);
	else	Verbose("CCXtable: none\n");

	if( CII_lsFonts(fonts,sizeof(fonts)) )
		Verbose("CIIfonts:\n%s",fonts);
	else	Verbose("CIIfonts: none\n");
}
